# -*- coding: utf-8 -*-
import numpy as np


class NegativeStdDevError(Exception):
    """An exception for when the standard deviation is negative"""

    pass


class NumpyDowncastWarning(RuntimeWarning):
    """An exception for when an uncertainties array is downcast to a numpy array"""

    pass


from .uncertainty import Uncertainty

try:
    from .pandas_compat import UncertaintyArray
except ImportError:
    pass


def nominal_values(x):
    # Is an Uncertainty
    if hasattr(x, "_nom"):
        return x.value
    else:
        if np.ndim(x) > 0:
            try:
                x2 = Uncertainty.from_sequence(x)
            except Exception:
                return x
            else:
                return x2.value
        else:
            try:
                x2 = Uncertainty(x)
            except Exception:
                return x
            else:
                return x2.value


def std_devs(x):
    # Is an Uncertainty
    if hasattr(x, "_nom"):
        return x.error
    else:
        if np.ndim(x) > 0:
            try:
                x2 = Uncertainty.from_sequence(x)
            except Exception:
                return np.zeros_like(x)
            else:
                return x2.error
        else:
            try:
                x2 = Uncertainty(x)
            except Exception:
                return 0
            else:
                return x2.value



# -*- coding: utf-8 -*-
from collections.abc import Iterable
import pandas as pd
import numpy as np
from pandas.api.extensions import ExtensionArray, ExtensionDtype, register_extension_dtype
from pandas.core.arrays.base import ExtensionOpsMixin
from pandas.compat import set_function_name
import loguru
import numbers

from .uncertainty import Uncertainty
from .util import is_np_duck_array, ignore_numpy_downcast_warnings


@register_extension_dtype
class UncertaintyType(ExtensionDtype):
    type = Uncertainty
    name = "uncertainty"

    def __init__(self):
        pass

    @property
    def _is_numeric(self):
        return True

    @property
    def _is_boolean(self):
        return False

    @classmethod
    def construct_array_type(cls):
        return UncertaintyArray

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))

    @property
    def na_value(self):
        return Uncertainty(np.nan, np.nan)

    @classmethod
    def construct_from_string(cls, string):
        if not isinstance(string, str):
            raise TypeError("'construct_from_string' expects a string, got {}".format(type(string)))
        elif string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from '{}'".format(cls.__name__, string))


class UncertaintyArray(ExtensionArray, ExtensionOpsMixin):
    # Required attributes
    @property
    def dtype(self):
        return UncertaintyType()

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def shape(self):
        return self.data.shape

    @property
    def nbytes(self):
        return self.data.nbytes

    # Required methods
    def __len__(self):
        return len(self.data)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __eq__(self, other):
        return self.data == other

    def isna(self):
        return np.isnan(self.data)

    def copy(self):
        inst = self.__class__(np.copy(self.data))
        return inst

    @ignore_numpy_downcast_warnings
    def take(self, indices, allow_fill=False, fill_value=None):
        from pandas.api.extensions import take

        if allow_fill:
            if fill_value is None or pd.isna(fill_value):
                fill_value = None
            elif not (isinstance(fill_value, Uncertainty) or fill_value is None):
                raise TypeError("Fill value must be Uncertainty or None")

        result = take(self.data, indices, allow_fill=allow_fill, fill_value=fill_value)
        if allow_fill and fill_value is None:
            result[pd.isna(result)] = None
        return UncertaintyArray(result)

    @classmethod
    def _concat_same_type(cls, to_concat):
        """
        Concatenate multiple array
        Parameters
        ----------
        to_concat : sequence of this type
        Returns
        -------
        ExtensionArray
        """
        data = np.concatenate([ga.data for ga in to_concat])
        return UncertaintyArray(data)

    def __getitem__(self, idx):
        if isinstance(idx, numbers.Integral):
            return self.data[idx]
        # array-like, slice
        # for pandas >= 1.0, validate and convert IntegerArray/BooleanArray
        # to numpy array, pass-through non-array-like indexers
        idx = pd.api.indexers.check_array_indexer(self, idx)
        if isinstance(idx, (Iterable, slice)):
            return UncertaintyArray(self.data[idx])
        else:
            raise TypeError("Index type not supported", idx)

    def _formatter(self, boxed=False):
        """Formatting function for scalar values.
        This is used in the default '__repr__'. The returned formatting
        function receives instances of your scalar type.
        Parameters
        ----------
        boxed: bool, default False
            An indicated for whether or not your array is being printed
            within a Series, DataFrame, or Index (True), or just by
            itself (False). This may be useful if you want scalar values
            to appear differently within a Series versus on its own (e.g.
            quoted or not).
        Returns
        -------
        Callable[[Any], str]
            A callable that gets instances of the scalar type and
            returns a string. By default, :func:`repr` is used
            when ``boxed=False`` and :func:`str` is used when
            ``boxed=True``.
        """
        return repr

    # def __repr__(self):
    #     print("B")
    #     return ""

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of scalars.
        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : boolean, default False
            If True, copy the underlying data.
        Returns
        -------
        ExtensionArray
        """
        return Uncertainty(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        """
        Reconstruct an ExtensionArray after factorization.
        Parameters
        ----------
        values : ndarray
            An integer ndarray with the factorized values.
        original : ExtensionArray
            The original ExtensionArray that factorize was called on.
        See Also
        --------
        pandas.factorize
        ExtensionArray.factorize
        """
        raise NotImplementedError

    def fillna(self, value=None, method=None, limit=None):
        """Fill NA/NaN values using the specified method.
        Parameters
        ----------
        value : scalar, array-like
            If a scalar value is passed it is used to fill all missing values.
            Alternatively, an array-like 'value' can be given. It's expected
            that the array-like have the same length as 'self'.
        method : {'backfill', 'bfill', 'pad', 'ffill', None}, default None
            Method to use for filling holes in reindexed Series
            pad / ffill: propagate last valid observation forward to next valid
            backfill / bfill: use NEXT valid observation to fill gap
        limit : int, default None
            If method is specified, this is the maximum number of consecutive
            NaN values to forward/backward fill. In other words, if there is
            a gap with more than this number of consecutive NaNs, it will only
            be partially filled. If method is not specified, this is the
            maximum number of entries along the entire axis where NaNs will be
            filled.
        Returns
        -------
        filled : ExtensionArray with NA/NaN filled
        """
        if method is not None:
            raise NotImplementedError("fillna with a method is not yet supported")

        mask = self.isna()
        new_values = self.copy()

        if mask.any():

            # fill with value
            if not isinstance(value, Uncertainty):
                raise NotImplementedError(
                    "fillna currently only supports filling with an Uncertainty"
                )
            new_values.data[mask] = value

        return new_values

    def __init__(self, value, error=None):

        if isinstance(value, Uncertainty):
            self.data = value
        else:
            value = np.asarray(value)

            if not is_np_duck_array(type(value)):
                raise TypeError("Values provided must be ndarrays.")
            if not value.ndim == 1:
                raise ValueError("Values provided must be a 1D array")
            if error is None:
                error = np.zeros_like(value)
            else:
                error = np.asarray(error)
            self.data = Uncertainty(value, error)

    @classmethod
    def _create_method(cls, op, coerce_to_dtype=True):
        """
        A class method that returns a method that will correspond to an
        operator for an ExtensionArray subclass, by dispatching to the
        relevant operator defined on the individual elements of the
        ExtensionArray.
        Parameters
        ----------
        op : function
            An operator that takes arguments op(a, b)
        coerce_to_dtype :  bool
            boolean indicating whether to attempt to convert
            the result to the underlying ExtensionArray dtype
            (default True)
        Returns
        -------
        A method that can be bound to a method of a class
        Example
        -------
        Given an ExtensionArray subclass called MyExtensionArray, use
        >>> __add__ = cls._create_method(operator.add)
        in the class definition of MyExtensionArray to create the operator
        for addition, that will be based on the operator implementation
        of the underlying elements of the ExtensionArray
        """

        def _binop(self, other):
            if isinstance(other, (pd.Series, pd.DataFrame)):
                return NotImplemented

            lvalues = self.data
            rvalues = other
            res = op(lvalues, rvalues)

            if op.__name__ == "divmod":
                return (cls(res[0]), cls(res[1]))

            if coerce_to_dtype:
                try:
                    res = cls(res)
                except TypeError:
                    pass

            return res

        op_name = f"__{op}__"
        return set_function_name(_binop, op_name, cls)

    @classmethod
    def _create_arithmetic_method(cls, op):
        return cls._create_method(op)

    @classmethod
    def _create_comparison_method(cls, op):
        return cls._create_method(op, coerce_to_dtype=False)


UncertaintyArray._add_arithmetic_ops()
UncertaintyArray._add_comparison_ops()



# -*- coding: utf-8 -*-
import warnings
from functools import wraps
from . import NumpyDowncastWarning


def ignore_runtime_warnings(f):
    """
    A decorator to ignore runtime warnings
    Parameters
    ----------
    f: function
        The wrapped function

    Returns
    -------
    wrapped_function: function
        The wrapped function
    """

    @wraps(f)
    def runtime_warn_inner(*args, **kwargs):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("ignore", category=RuntimeWarning)
            response = f(*args, **kwargs)
        return response

    return runtime_warn_inner


def ignore_numpy_downcast_warnings(f):
    """
    A decorator to ignore NumpyDowncastWarning
    Parameters
    ----------
    f: function
        The wrapped function

    Returns
    -------
    wrapped_function: function
        The wrapped function
    """

    @wraps(f)
    def user_warn_inner(*args, **kwargs):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("ignore", category=NumpyDowncastWarning)
            response = f(*args, **kwargs)
        return response

    return user_warn_inner


def is_iterable(y):
    try:
        iter(y)
    except TypeError:
        return False
    return True


def has_length(y):
    try:
        len(y)
    except TypeError:
        return False
    return True


def is_np_duck_array(cls):
    """Check if object is a numpy array-like, but not a Uncertainty

    Parameters
    ----------
    cls : class

    Returns
    -------
    bool
    """
    try:
        import numpy as np
    except ImportError:
        return False

    return issubclass(cls, np.ndarray) or (
        not hasattr(cls, "_nom")
        and not hasattr(cls, "_err")
        and hasattr(cls, "__array_function__")
        and hasattr(cls, "ndim")
        and hasattr(cls, "dtype")
    )


class Display(object):
    default_format: str = ""

    def _repr_html_(self):
        if hasattr(self._nom, "units"):
            val_ = self._nom.m
            err_ = self._err.m
            u = self._nom.units
        else:
            val_ = self._nom
            err_ = self._err
            u = None
        if is_np_duck_array(type(self._nom)):
            header = "<table><tbody>"
            footer = "</tbody></table>"
            val = f"<tr><th>Magnitude</th><td style='text-align:left;'><pre>{val_}</pre></td></tr>"
            err = f"<tr><th>Error</th><td style='text-align:left;'><pre>{err_}</pre></td></tr>"
            if u is None:
                units = ""
            else:
                units = (
                    f"<tr><th>Units</th><td style='text-align:left;'>{u._repr_html_()}</td></tr>"
                )
            return header + val + err + units + footer
        else:
            val = f"{val_}"
            err = f"{err_}"
            if u is None:
                units = ""
            else:
                units = f"{u}"
            return f"{val} {chr(0x00B1)} {err} {units}"

    def _repr_latex_(self):
        if hasattr(self._nom, "units"):
            val_ = self._nom.m
            err_ = self._err.m
            u = self._nom.units
        else:
            val_ = self._nom
            err_ = self._err
            u = None
        if is_np_duck_array(type(self._nom)):
            s = ", ".join([f"{v} \\pm {e}" for v, e in zip(val_.ravel(), err_.ravel())]) + "~"
            header = "$"
            footer = "$"
            if u is None:
                units = ""
            else:
                units = u._repr_latex_()
            return header + s + units + footer
        else:
            val = f"{val_}"
            err = f"{err_}"
            if u is None:
                units = ""
            else:
                units = u._repr_latex_()
            return f"${val} \\pm {err} {units}$"

    def __str__(self) -> str:
        if hasattr(self._nom, "units"):
            val_ = self._nom.m
            err_ = self._err.m
            u = self._nom.units
        else:
            val_ = self._nom
            err_ = self._err
            u = None

        if u is None:
            units = ""
        else:
            units = f" {u}"

        if self._nom is not None:
            if self._err is not None:
                if is_np_duck_array(type(self._nom)):
                    return (
                        "["
                        + ", ".join([f"{v} +/- {e}" for v, e in zip(val_.ravel(), err_.ravel())])
                        + "]"
                        + units
                    )
                else:
                    return f"{self._nom} +/- {self._err}" + units
            else:
                return f"{self._nom}" + units

    def __format__(self, fmt):
        return f"{self.value:{fmt}} +/- {self.error:{fmt}}"

    def __repr__(self) -> str:
        return str(self)



# -*- coding: utf-8 -*-
# Based heavily on the implementation of pint's Quantity object
from __future__ import annotations
from types import MethodType

import numpy as np
import locale
import copy
import operator
import warnings
import jax
from pint import Quantity, DimensionalityError
from pint.util import SharedRegistryObject

from .wrap_numpy import wrap_numpy, HANDLED_FUNCTIONS, HANDLED_UFUNCS
from . import NegativeStdDevError, NumpyDowncastWarning
from .util import is_np_duck_array, ignore_runtime_warnings, ignore_numpy_downcast_warnings, Display


def _check_units(value, err):
    mag_units = hasattr(value, "units")
    err_units = hasattr(err, "units")
    if mag_units ^ err_units and err is not None:
        raise ValueError("Both mag and err need to have units if one of them has units!")
    if mag_units and err is not None:
        if value.units != err.units:
            raise ValueError(
                f"Value units {value.units} cannot be converted to error units {err.units}"
            )
    if mag_units:
        mag_units = value.units
        ret_val = value.to(mag_units).m
        if err is not None:
            ret_err = err.to(mag_units).m
        else:
            ret_err = None
    else:
        mag_units = 1.0
        ret_val = value
        ret_err = err

    return ret_val, ret_err, mag_units


def _strip_device_array(value, err):
    if isinstance(value, jax.xla.DeviceArray):
        value = value.to_py().copy()
    if isinstance(err, jax.xla.DeviceArray):
        err = err.to_py().copy()
    return value, err


class Uncertainty(Display):
    __apply_to_both_ndarray__ = ["flatten", "real", "imag", "astype", "T"]
    __ndarray_attributes__ = ["dtype", "ndim", "size"]

    # Pint comparibility
    @property
    def unitless(self) -> bool:
        """ """
        return not bool(self.to_root_units()._units)

    @property
    def dimensionless(self) -> bool:
        """ """
        tmp = self.to_root_units()

        return not bool(tmp.dimensionality)

    _dimensionality = None

    @property
    def dimensionality(self):
        """
        Returns
        -------
        dict
            Dimensionality of the Quantity, e.g. ``{length: 1, time: -1}``
        """
        if self._dimensionality is None:
            self._dimensionality = self._REGISTRY._get_dimensionality(self.units)

        return self._dimensionality

    def check(self, dimension) -> bool:
        """Return true if the quantity's dimension matches passed dimension."""
        return self.dimensionality == self._REGISTRY.get_dimensionality(dimension)

    def _check(self, other) -> bool:
        """Check if the other object use a registry and if so that it is the
        same registry.
        Parameters
        ----------
        other :
        Returns
        -------
        type
            other don't use a registry and raise ValueError if other don't use the
            same unit registry.
        """
        if self._REGISTRY is getattr(other, "_REGISTRY", None):
            return True

        elif isinstance(other, SharedRegistryObject):
            mess = "Cannot operate with {} and {} of different registries."
            raise ValueError(mess.format(self.__class__.__name__, other.__class__.__name__))
        else:
            return False

    def to(self, other):
        if hasattr(self.nominal_value, "units"):
            return self.__class__(self._nom.to(other), self._err.to(other))
        else:
            raise AttributeError("Uncertainty has no quantity!")

    @property
    def m(self):
        if hasattr(self.nominal_value, "units"):
            return self.__class__(self._nom.m, self._err.m)
        else:
            raise AttributeError("Uncertainty has no quantity!")

    @property
    def units(self):
        if hasattr(self.nominal_value, "units"):
            return self._nom.units
        else:
            raise AttributeError("Uncertainty has no quantity!")

    @property
    def _REGISTRY(self):
        return getattr(self._nom, "_REGISTRY", None)

    def compatible_units(self, *contexts):
        if contexts:
            with self._REGISTRY.context(*contexts):
                return self._REGISTRY.get_compatible_units(self._units)

        return self._REGISTRY.get_compatible_units(self._units)

    def _convert_magnitude_not_inplace(self, other, *contexts, **ctx_kwargs):
        if contexts:
            with self._REGISTRY.context(*contexts, **ctx_kwargs):
                return self._REGISTRY.convert(self._magnitude, self._units, other)

        return self._REGISTRY.convert(self._magnitude, self._units, other)

    @ignore_numpy_downcast_warnings
    def __init__(self, value, err=None):

        value, err, units = _check_units(value, err)
        value, err = _strip_device_array(value, err)

        # If Uncertatity
        if isinstance(value, self.__class__):
            magnitude_nom = value.value
            magnitude_err = value.error
        # If sequence
        elif isinstance(value, list):
            return self.__class__.from_list(value)
        # If arrays
        elif np.ndim(value) > 0:
            magnitude_nom = np.asarray(value)
            if err is None:
                magnitude_err = np.zeros_like(value)
            else:
                if np.ndim(err) == 0:
                    magnitude_err = np.ones_like(value) * err
                else:
                    magnitude_err = np.asarray(err)
                    assert magnitude_err.shape == magnitude_nom.shape
        # If scalar
        else:
            magnitude_nom = value
            if err is None:
                magnitude_err = 0.0
            else:
                magnitude_err = err

        # Replace NaNs in errors with zeros
        if is_np_duck_array(type(magnitude_err)):
            magnitude_err[~np.isfinite(magnitude_err)] = 0
        else:
            if not np.isfinite(magnitude_err):
                magnitude_err = 0
        magnitude_nom *= units
        magnitude_err *= units
        # Basic sanity checks
        if is_np_duck_array(type(magnitude_nom)):
            for item in self.__ndarray_attributes__ + ["shape"]:
                if not getattr(magnitude_nom, item) == getattr(magnitude_err, item):
                    raise ValueError(
                        f"Attribute {item} does not match for value and error! ({getattr(magnitude_nom,item)} and {getattr(magnitude_err,item)})"
                    )
        err_mag = np.atleast_1d(magnitude_err)
        if np.any(err_mag[np.isfinite(err_mag)] < 0):
            valid = err_mag[np.isfinite(err_mag)]

            raise NegativeStdDevError(
                f"Found negatives values for the standard deviation... {valid[valid < 0]}"
            )

        self._nom = magnitude_nom
        self._err = magnitude_err

    def __bytes__(self) -> bytes:
        return str(self).encode(locale.getpreferredencoding())

    def __iter__(self):
        for v, e in zip(self._nom, self._err):
            yield self.__class__(v, e)

    def __copy__(self) -> Uncertainty:
        ret = self.__class__(copy.copy(self._nom), copy.copy(self._err))

        return ret

    def __deepcopy__(self, memo) -> Uncertainty:
        ret = self.__class__(copy.deepcopy(self._nom, memo), copy.deepcopy(self._err, memo))
        return ret

    def __hash__(self) -> int:
        return hash((self.__class__, self._nom, self._err))

    @property
    def value(self):
        return self._nom

    @property
    def nominal_value(self):
        return self.value

    @property
    def n(self):
        return self.value

    @property
    def error(self):
        return self._err

    @property
    def std_dev(self):
        return self.error

    @property
    def s(self):
        return self.error

    @property
    def relative(self):
        if np.ndim(self._nom) == 0:
            if self._nom != 0:
                return self._err / self._nom
            else:
                return np.NaN
        else:
            return self._err / self._nom

    @property
    def rel(self):
        return self.relative

    @classmethod
    def from_list(cls, u_list):
        return cls.from_sequence(u_list)

    @classmethod
    def from_sequence(cls, seq):
        _ = iter(seq)

        len_seq = len(seq)
        val = np.empty(len_seq)
        err = np.empty(len_seq)
        if len_seq > 0:
            first_item = seq[0]
            try:
                first_item + 1
            except TypeError:
                raise TypeError(
                    f"Sequence elements of type {type(first_item)} dont support math operations!"
                )
            if hasattr(first_item, "units"):
                val *= first_item.units
                err *= first_item.units
            for i, seq_i in enumerate(seq):
                try:
                    val[i] = float(seq_i._nom)
                    err[i] = float(seq_i._err)
                except AttributeError:
                    val[i] = float(seq_i)
                    err[i] = 0

        return cls(val, err)

    def __float__(self) -> Uncertainty:
        return float(self._nom)

    def __complex__(self) -> Uncertainty:
        return complex(self._nom)

    def __int__(self) -> Uncertainty:
        return int(self._nom)

    # Math Operators
    def __iadd__(self, other):
        new = self + other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    def __add__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom + other._nom
            new_err = np.sqrt(self._err ** 2 + other._err ** 2)
        else:
            new_mag = self._nom + other
            new_err = self._err
        return self.__class__(new_mag, new_err)

    __radd__ = __add__

    def __isub__(self, other):
        new = self - other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    def __sub__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom - other._nom
            new_err = np.sqrt(self._err ** 2 + other._err ** 2)
        else:
            new_mag = self._nom - other
            new_err = self._err
        return self.__class__(new_mag, new_err)

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __imul__(self, other):
        new = self * other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    def __mul__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom * other._nom
            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
        else:
            new_mag = self._nom * other
            new_err = np.abs(self._err * other)

        return self.__class__(new_mag, new_err)

    __rmul__ = __mul__

    @ignore_runtime_warnings
    def __itruediv__(self, other):
        new = self / other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    @ignore_runtime_warnings
    def __truediv__(self, other):
        # Self / Other
        if isinstance(other, Uncertainty):
            new_mag = self._nom / other._nom
            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
        else:
            new_mag = self._nom / other
            new_err = np.abs(self._err / other)
        return self.__class__(new_mag, new_err)

    @ignore_runtime_warnings
    def __rtruediv__(self, other):
        # Other / Self
        if isinstance(other, Uncertainty):
            raise Exception
        else:
            new_mag = other / self._nom
            new_err = np.abs(new_mag) * np.abs(self.rel)
            return self.__class__(new_mag, new_err)

    __div__ = __truediv__
    __rdiv__ = __rtruediv__
    __idiv__ = __itruediv__

    def __ifloordiv__(self, other):
        new = self // other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    def __floordiv__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom // other._nom
            new_err = 0.0
        else:
            new_mag = self._nom // other
            new_err = 0.0
        return self.__class__(new_mag, new_err)

    def __rfloordiv__(self, other):
        if isinstance(other, Uncertainty):
            return other.__truediv__(self)
        else:
            new_mag = other // self._nom
            new_err = 0.0
            return self.__class__(new_mag, new_err)

    def __imod__(self, other):
        new = self % other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    def __mod__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom % other._nom
        else:
            new_mag = self._nom % other
        if np.ndim(new_mag) == 0:
            new_err = 0.0
        else:
            new_err = np.zeros_like(new_mag)
        return self.__class__(new_mag, new_err)

    def __rmod__(self, other):
        new_mag = other % self._nom
        if np.ndim(new_mag) == 0:
            new_err = 0.0
        else:
            new_err = np.zeros_like(new_mag)
        return self.__class__(new_mag, new_err)

    def __divmod__(self, other):
        return self // other, self % other

    def __rdivmod__(self, other):
        return other // self, other % self

    def __ipow__(self, other):
        new = self ** other
        if is_np_duck_array(type(self._nom)):
            self._err = new._err
            self._nom = new._nom
            return self
        else:
            return new

    @ignore_runtime_warnings
    def __pow__(self, other):
        # Self ** other
        A = self._nom
        sA = self._err
        if isinstance(other, Uncertainty):
            B = other._nom
            sB = other._err
        else:
            B = other
            sB = 0
        new_mag = A ** B
        if sB == 0 and int(B) == B:
            new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2)
        else:
            new_err = np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)

        return self.__class__(new_mag, new_err)

    @ignore_runtime_warnings
    def __rpow__(self, other):
        # Other ** self
        B = self._nom
        sB = self._err
        if isinstance(other, Uncertainty):
            A = other._nom
            sA = other._err
        else:
            A = other
            sA = 0

        new_mag = A ** B
        new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)

        return self.__class__(new_mag, new_err)

    def __abs__(self):
        return self.__class__(abs(self._nom), self._err)

    def __round__(self, ndigits):
        return self.__class__(round(self._nom, ndigits=ndigits), self._err)

    def __pos__(self):
        return self.__class__(operator.pos(self._nom), self._err)

    def __neg__(self):
        return self.__class__(operator.neg(self._nom), self._err)
        return self.__class__(operator.neg(self._nom), self._err)

    def __eq__(self, other):
        if isinstance(other, Uncertainty):
            return self._nom == other._nom
        else:
            return self._nom == other

    def __ne__(self, other):
        out = self.__eq__(other)
        if is_np_duck_array(type(out)):
            return np.logical_not(out)
        else:
            return not out

    def compare(self, other, op):
        if isinstance(other, Uncertainty):
            return op(self._nom, other._nom)
        else:
            return op(self._nom, other)

    __lt__ = lambda self, other: self.compare(other, op=operator.lt)
    __le__ = lambda self, other: self.compare(other, op=operator.le)
    __ge__ = lambda self, other: self.compare(other, op=operator.ge)
    __gt__ = lambda self, other: self.compare(other, op=operator.gt)

    def __bool__(self) -> bool:
        return bool(self._nom)

    __nonzero__ = __bool__

    # NumPy function/ufunc support
    __array_priority__ = 17

    @ignore_runtime_warnings
    def __array_function__(self, func, types, args, kwargs):
        # print(func)
        if func.__name__ not in HANDLED_FUNCTIONS:
            return NotImplemented
        elif not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        else:
            return wrap_numpy("function", func, args, kwargs)

    @ignore_runtime_warnings
    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        # print(method,ufunc.__name__)
        if method != "__call__":
            raise NotImplementedError
        else:
            if ufunc.__name__ not in HANDLED_UFUNCS:
                raise NotImplementedError(f"Ufunc {ufunc.__name__} is not implemented!") from None
            else:
                return wrap_numpy("ufunc", ufunc, args, kwargs)

    def __getattr__(self, item):
        if item.startswith("__array_"):
            # Handle array protocol attributes other than `__array__`
            raise AttributeError(f"Array protocol attribute {item} not available.")
        elif item in self.__apply_to_both_ndarray__:
            val = getattr(self._nom, item)
            err = getattr(self._err, item)

            if callable(val):
                return lambda *args, **kwargs: self.__class__(
                    val(*args, **kwargs), err(*args, **kwargs)
                )
            else:
                return self.__class__(val, err)
        elif item in HANDLED_UFUNCS:
            return lambda *args, **kwargs: wrap_numpy("ufunc", item, [self] + list(args), kwargs)
        elif item in HANDLED_FUNCTIONS:
            return lambda *args, **kwargs: wrap_numpy("function", item, [self] + list(args), kwargs)
        elif item in self.__ndarray_attributes__:
            return getattr(self._nom, item)
        elif hasattr(Quantity, item):
            val = getattr(self._nom, item)
            err = getattr(self._err, item)
            if callable(val):

                def expr(*args, **kwargs):
                    try:
                        vexpr = val(*args, **kwargs)
                        eexpr = err(*args, **kwargs)
                    except Exception:
                        raise ValueError(
                            f"Could not execute method {item} on Uncertainty elements!"
                        )
                    try:
                        ret_instance = self.__class__(vexpr, eexpr)
                    except Exception:
                        raise ValueError(
                            f"Could not execute instantiate Uncertainty class with results from method {item}!"
                        )
                    else:
                        return ret_instance

                return expr
            else:
                return self.__class__(val, err)
        else:
            raise AttributeError(
                f"Attribute {item} not available in Uncertainty, as method of a Pint Quantity, or as NumPy ufunc or function."
            ) from None

    def __array__(self, t=None) -> np.ndarray:
        warnings.warn(
            "The uncertainty is stripped when downcasting to ndarray.",
            NumpyDowncastWarning,
            stacklevel=2,
        )
        return np.asarray(self._nom)

    def clip(self, min=None, max=None, out=None, **kwargs):

        return self.__class__(self._nom.clip(min, max, out, **kwargs), self._err)

    def fill(self, value) -> None:
        return self._nom.fill(value)

    def put(self, indices, values, mode="raise") -> None:
        if isinstance(values, self.__class__):
            self._nom.put(indices, values._nom, mode)
            self._err.put(indices, values._err, mode)
        else:
            raise ValueError("Can only 'put' Uncertainties into uncertainties!")

    def copy(self):
        return Uncertainty(self._nom.copy(), self._err.copy())

    # Special properties
    @property
    def flat(self):
        for u, v in (self._nom.flat, self._err.flat):
            yield self.__class__(u, v)

    @property
    def shape(self):
        return self._nom.shape

    @shape.setter
    def shape(self, value):
        self._nom.shape = value
        self._err.shape = value

    @property
    def nbytes(self):
        return self._nom.nbytes + self._err.nbytes

    def searchsorted(self, v, side="left", sorter=None):
        return self._nom.searchsorted(v, side)

    def __len__(self) -> int:
        return len(self._nom)

    def __getitem__(self, key):
        try:
            return self.__class__(self._nom[key], self._err[key])
        except TypeError:
            raise TypeError(f"Index {key} not supported!")

    def __setitem__(self, key, value):
        if not isinstance(value, self.__class__):
            raise ValueError(
                f"Can only pass Uncertainty type to __setitem__! Instead passed {type(value)}"
            )
        try:
            _ = self._nom[key]
        except ValueError as exc:
            raise ValueError(f"Object {type(self._nom)} does not support indexing") from exc

        self._nom[key] = value._nom
        self._err[key] = value._err

    def tolist(self):
        try:
            nom = self._nom.tolist()
            err = self._err.tolist()
            if not isinstance(nom, list):
                return self.__class__(nom, err)
            else:
                return [
                    self.__class__(n, e).tolist() if isinstance(n, list) else self.__class__(n, e)
                    for n, e in (nom, err)
                ]
        except AttributeError:
            raise AttributeError(f"{type(self._nom).__name__}' does not support tolist.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['auto_uncertainties']

package_data = \
{'': ['*']}

install_requires = \
['jax', 'numpy>1.20']

setup_kwargs = {
    'name': 'auto-uncertainties',
    'version': '0.3.1',
    'description': 'Linear Uncertainty Propagation with Auto-Differentiation',
    'long_description': None,
    'author': 'Varchas Gopalaswamy',
    'author_email': 'vgop@lle.rochester.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
